--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- Console version 1.0.8 (21.7.2013)
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- A fairly simple in-game console implementation in Love2D. Contains support
-- for input history, scrolling text, and basic interpretation at run-time.
--------------------------------------------------------------------------------
-- Notes:
-- Since the console is mainly used for debugging there isn't much optimization.
--
-- Not really happy with the way it is drawn. So many tedious, per-pixel
-- calculations. Also no support for color tags or anything fancy :(. I would
-- like to improve it, but it would probably mean keeping coordinates for each
-- word (or even char) of text and printing them individually. Someday...
--
--------------------------------------------------------------------------------
-- Useage:
-- Simply require("console") to get the console module. Next, be sure to call
-- console.load() in love.love(), console.draw() in love.draw(), and
-- console.keypressed() in love.keypressed(). Edit the settings below to taste.
-- Oh, and make sure you have the ttf file for the font you wish to use.
--------------------------------------------------------------------------------
-- The MIT License (MIT)
--
-- Copyright (c) 2013 Nicholas Campbell
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Module definition / settings
--------------------------------------------------------------------------------
local console = {}

console.fontFile = "Consolas.ttf"
console.fontSize = 12
console.textColor = {255, 255, 255, 190}
console.lineSpacing = 1
console.font = love.graphics.getFont()

console.backgroundColor = {0, 0, 0, 50}

console.optionBar = true
console.optionBarColor = { 255, 255, 255, 190}

console.inputBorderSize = 3
console.inputBorderColor = {255, 0, 255, 90}
console.prompt = "> "
console.promptIndent = 5
console.inputColor = {255, 0, 255, 50}
console.inputBuffer = ""
console.keyRepeatDelay = 0.20
console.keyRepeatInterval = 0.05
console.echoInput = true

console.historyBuffer = {}
console.historyPosition = 0

console.cursorColor = {255, 0, 255, 255}
console.cursorBlinkRate = 0.6
console.cursorPosition = 0

console.buffer = {}
console.lines = 20
console.linePosition = 0
console.textboxBB = {["x"] = 0, ["y"] = 0, ["w"] = 0, ["h"] = 0}

console.activateKey = "`"
console.active = false

--------------------------------------------------------------------------------
-- Utility functions
--------------------------------------------------------------------------------
local console_utils = require("console_util")

local show = console_utils.show
local pack = console_utils.pack
local modulo = console_utils.modulo
local round = console_utils.round
local insert_char = console_utils.insert_char
local replace_char = console_utils.replace_char
local split = console_utils.split

--------------------------------------------------------------------------------
-- Local functions
--------------------------------------------------------------------------------
local function addText(str)
    local t = {}
    
    if str == nil then str = "nil" end
    
    -- Split on newline; our buffer is a table list of strings and each index is a line
    t = split(str, '\n')
    
    -- Merge text
    for i = 1, #t do
        console.buffer[#console.buffer + 1] = t[i]
    end
    
    console.linePosition = #console.buffer
end

local function drawVisibleText()
    if console.linePosition == 0 then return end
    
    local buff = console.buffer
    
    local lineHeight = console.lineSpacing + console.font:getHeight()
    local y = console.textboxBB.h -  lineHeight
    
    -- Determine endpoint; ie., how "low" into the buffer we might possibly need
    -- to go to fill the textbox using console.lines as the number of lines
    local endpoint = console.linePosition - console.lines
    if endpoint < 1 then endpoint = 1 end
    
    -- Draw each line in the buffer inside the range created by console.lines
    for line = console.linePosition, endpoint, -1 do
        if console.font:getWidth(buff[line]) <= love.graphics.getWidth() then
            -- Not going to be wrapped; simply print it at the current Y position
            love.graphics.print(buff[line], 0, y)
        else
            -- Text is gonna wrap; figure out how many lines and draw at that position
            local wrapWidth, wrapCount = console.font:getWrap(buff[line], love.graphics.getWidth())
            y = y - (lineHeight * (wrapCount - 1))
            love.graphics.printf(buff[line], 0, y, love.graphics.getWidth())           
        end
                    
        -- Increment Y
        y = y - lineHeight
    end
end

local function calculateTextboxBB()
    -- The bounding box for the text area. Height is simply the number of lines times the
    -- height of the font, plus any line spacing.
    console.textboxBB.x = 0
    console.textboxBB.y = 0
    console.textboxBB.w = love.graphics.getWidth()
    console.textboxBB.h = (console.lines * console.font:getHeight()) + (console.lines * console.lineSpacing)
end

local function drawTextbox()
    -- Preserve previous color
    local oldColor = pack(love.graphics.getColor())
    
    -- Recalc dimensions of textbox
    calculateTextboxBB()
    local bb = console.textboxBB
    
    -- Draw bg
    love.graphics.setColor(console.backgroundColor)
    love.graphics.rectangle("fill", bb.x, bb.y, bb.w, bb.h)
    
    -- Draw text
    love.graphics.setColor(console.textColor)
    drawVisibleText()

    -- Return to previous color state
    love.graphics.setColor(oldColor)
end

local function drawBorder(y, w, borderSize)
    -- Draw the line based on the border width
    -- https://love2d.org/forums/viewtopic.php?f=4&t=10464&p=63090
    love.graphics.setLineWidth(borderSize)
    love.graphics.setColor(console.inputBorderColor)
    love.graphics.line(0, y + (borderSize / 2), w, y + (borderSize / 2))
    love.graphics.setLineWidth(1)
end

local function drawInputArea()
    -- Preserve color info
    local oldColor = pack(love.graphics.getColor())
    
    -- Info used for drawing calls. Width is always the width of the window and y is
    -- set to the height of the console's textbox area (right after where we want to draw
    -- the borders and input area).
    local y, w = console.textboxBB.h, love.graphics.getWidth()
    local borderSize = console.inputBorderSize
    local lineSpacing = console.lineSpacing

    -- Draw top border after the textbox area
    drawBorder(y, w, borderSize)
    
    -- Draw input background; height is the height of the font plus the padding
    -- from the line spacing
    local bgHeight = console.font:getHeight() + lineSpacing
    love.graphics.setColor(console.backgroundColor)
    love.graphics.rectangle("fill", 0, y + borderSize, w, bgHeight)
    
    -- Draw prompt
    local promptWidth = console.promptIndent + console.font:getWidth(console.prompt) 
    love.graphics.setColor(console.inputBorderColor)
    love.graphics.print(console.prompt, 0 + console.promptIndent, y + borderSize + lineSpacing)
    
    -- Prepare to draw input text; special case when it starts to flow offscreen
    love.graphics.setColor(console.textColor)
    local txt = console.inputBuffer
    if (promptWidth + console.font:getWidth(console.inputBuffer)) > w then
        -- Input text has to scroll; shrink the text until we get a string that will fit
        local pos = 1
        while (promptWidth + console.font:getWidth(txt)) > w do
            txt = string.sub(console.inputBuffer, pos)
            pos = pos + 1
        end
    end
    
    -- Draw input text
    love.graphics.print(txt, 0 + promptWidth, y + borderSize + lineSpacing)
    
    -- First, a quick and dirty hack to handle cursor blinking: basically, taking the modulus
    -- of the microtimer and the blink rate will return a float in the range of ~0 to ~blinkrate.
    -- So, if the float returned is on the bottom half of the blink rate we draw it, else not :)
    if modulo(love.timer.getTime(), console.cursorBlinkRate) <= (console.cursorBlinkRate / 2) then
        -- Draw cursor; x position is calculated using the cursor's position and the width of the
        -- string up until this position. Then the prompt width is added because of its offset.
        love.graphics.setColor(console.cursorColor)
        local cursorX = console.font:getWidth(string.sub(txt, 1, console.cursorPosition))
        local cursorY = y + borderSize + lineSpacing
        cursorX = promptWidth + cursorX
        love.graphics.line(cursorX, cursorY, cursorX, cursorY + console.font:getHeight())
    end
    
    -- Draw bottom border after input area; y set past the top border and input area
    drawBorder(y + borderSize + bgHeight, w, borderSize)
    
    -- Draw option bar
    if console.optionBar == true then
        y = y + (borderSize * 2) + bgHeight
        love.graphics.setColor(console.inputBorderColor)
        love.graphics.rectangle("fill", 0, y, w, bgHeight)
        
        -- Draw optional info
        love.graphics.setColor(console.optionBarColor)
        local optionText = "FPS: " .. love.timer.getFPS() .. " | dt: " .. round(love.timer.getDelta(), 2)
        love.graphics.print(optionText, 0, y)
    end
    
    -- Return to previous color state
    love.graphics.setColor(oldColor)
end

local function updateCursorPosition(pos)
    -- Prevent going past input text
    if pos > 0 then
        if console.cursorPosition == string.len(console.inputBuffer) then return end
    end
    
    -- Prevent negative cursor position
    if pos < 0 then
        if console.cursorPosition == 0 then return end
    end

    console.cursorPosition = console.cursorPosition + pos
end

local function updateHistoryPosition(pos)
    if (console.historyPosition + pos) < 1 then return end
    if (console.historyPosition + pos) > #console.historyBuffer then return end
    
    console.historyPosition = console.historyPosition + pos
end

local function clearInputBuffer()
    console.inputBuffer = ""
    console.cursorPosition = 0
end

local function processInputEntry()
    -- Save cmd to history and return history position to +1 past the last;
    -- may seem weird to keep an index out of bounds, but when you think about
    -- it only once you push the "up" key to scroll through the history
    -- does the indexing start.
    console.historyBuffer[#console.historyBuffer + 1] = console.inputBuffer
    console.historyPosition = #console.historyBuffer + 1
    
    -- Echo input
    if console.echoInput == true then
        addText("> " .. console.inputBuffer)
    end
    
    -- Do shit based on cmd eventually; for now, just eval simple expressions
    -- TODO: add more functionality
    -- First, try it as though it were an expression (with return)
    local func, err = loadstring("return " .. console.inputBuffer)
    
    -- If that didn't work, try as a statement on its own
    local statement = false
    if func == nil then 
        func, err = loadstring(console.inputBuffer) 
        statement = true
    end
    
    -- Function is constructed, use pcall if its valid or catch any errors
    if func ~= nil then
        local status, result = pcall(func)
        
        -- For statements, we acknowledge the statement; anything else display result
        if statement then
            addText("Ok.")
        else        
            addText(show(result))
        end
    else
        addText(err)
    end
    
    -- Clear input buffer
    clearInputBuffer()
end

local function scrollText(dir)
    console.linePosition = console.linePosition + (dir * console.lines)
  
    if console.linePosition < console.lines then console.linePosition = console.lines end
    if console.linePosition > #console.buffer then console.linePosition = #console.buffer end
end

--------------------------------------------------------------------------------
-- Module functions
--------------------------------------------------------------------------------
function console.load()
    console.font = love.graphics.newFont(console.fontFile, console.fontSize)
    
    love.keyboard.setKeyRepeat(console.keyRepeatDelay, console.keyRepeatInterval)
end

function console.draw()
    if console.active == false then return end
    
    -- Preserve old font and enable console's
    local oldFont = love.graphics.getFont()
    love.graphics.setFont(console.font)
    
    drawTextbox()
    drawInputArea()
    
    -- Return to previous font state
    if oldFont then love.graphics.setFont(oldFont) end
end

function console.keypressed(key, unicode)
    local inputBuffer = console.inputBuffer
    local cursorPos = console.cursorPosition
    
    -- Check for activation key
    if key == console.activateKey then
        console.active = not console.active
        return
    end
    
    if key == "kp0" then
        addText(show(upvalues()))
    end
    
    -- We only want to process key presses if the console is active
    if console.active == false then return end

    -- Deleting input text
    if key == "backspace" and console.cursorPosition ~= 0 then
        console.inputBuffer = replace_char(inputBuffer, cursorPos, "")
        updateCursorPosition(-1)
    elseif key == "delete" then
        console.inputBuffer = replace_char(inputBuffer, cursorPos + 1, "")   
    end
    
    -- Manipulating the cursor
    if key == "right" then
        updateCursorPosition(1)
    elseif key == "left" then
        updateCursorPosition(-1)
    elseif key == "home" then
        console.cursorPosition = 0
    elseif key == "end" then
        console.cursorPosition = string.len(console.inputBuffer)
    end
    
    -- Submitting input text
    if key == "return" then
        processInputEntry()
    end
    
    -- Scrolling through history
    if key == "up" or key == "down" then
        if key == "up" then updateHistoryPosition(-1) end
        if key == "down" then updateHistoryPosition(1) end
        
        clearInputBuffer()
        
        -- Display the command in history (if available)
        if console.historyPosition <= #console.historyBuffer and
           #console.historyBuffer > 0 then
            console.inputBuffer = console.historyBuffer[console.historyPosition]
            console.cursorPosition = #console.inputBuffer
        end
    end
    
    -- Page scrolling
    if key == "pageup" then
        scrollText(-1)
    elseif key == "pagedown" then
        scrollText(1)
    end
    
    -- Handle remaining unicode strings
    if unicode then
        -- 0.9.0
        if love._version >= "0.9.0" then
            -- Convert to bytecode (integer)
            unicode = string.byte(unicode, 1)
        end
        
        -- Took this from the love.keypressed wiki entry...seems to work for English
        -- YMMV with other languages / crazy unicodeness
        if unicode > 31 and unicode < 127 then
            console.inputBuffer = insert_char(inputBuffer, cursorPos, string.char(unicode))
            updateCursorPosition(1)
        end
    end
end

return console